////////////////////////////////////////////////////////////////////// 
// fugue/HaploSet.cpp 
// (c) 2000-2003 Goncalo Abecasis
// 
// This file is distributed as part of the FUGUE source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile FUGUE.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Tuesday July 06, 2004
// 
 
#include "HaploSet.h"
#include "Error.h"
#include "StringArray.h"

#include <math.h>

int HaplotypeSet::maxHaploCount = 0;

HaplotypeSet::HaplotypeSet()
   {
   size = haploCount = graphCount = 0;
   graphs = NULL;
   weight = 0.0;
   }

HaplotypeSet::~HaplotypeSet()
   {
   if (graphs != NULL)
      {
      for (int i = 0; i < graphCount; i++)
         delete graphs[i];
      delete [] graphs;
      }
   }

void HaplotypeSet::Grow()
   {
   size = size == 0 ? 1 : size * 2;

   HaplotypeGraph ** new_graphs = new HaplotypeGraph * [size];

   for (int i = 0; i < graphCount; i++)
      new_graphs[i] = graphs[i];

   if (graphs != NULL) delete [] graphs;

   graphs = new_graphs;
   }

// get founder haplotypes sets for a family
void HaplotypeSet::GetHaplotypeSets(
                     String* haps, int haplos, int wgt)
   {
   HaplotypeGraph * graph = new HaplotypeGraph;

   if (haploCount != 0 && haploCount != haplos)
      error("Expecting %d Haplotypes in Family %d\n",
            haploCount, (const char *) setid);

   haploCount = haplos;
   if (haplos > maxHaploCount)
      maxHaploCount = haploCount;

   graph->GetHaplotypes(haps, haplos);
   weight += graph->weight = wgt;

   if (graphCount >= size) Grow();

   graphs[graphCount++] = graph;
   }

void HaplotypeSet::StandardizeWeights()
   {
   if (weight != 1.0)
      {
      for (int i = 0; i < graphCount; i++)
         graphs[i]->weight /= weight;
      weight = 1.0;
      }
   }


double HaplotypeSet::Complexity()
   {
   double c = 0;

   for (int i = 0; i < graphCount; i++)
      c += graphs[i]->Complexity();

   return c;
   }

void HaplotypeSet::UpdateAlleleCounts()
   {
   for (int i = 0; i < graphCount; i++)
      graphs[i]->UpdateAlleleCounts();
   }

// set marker count
void HaplotypeSets::SetMarkerCount(int marker_no){
   MarkerInfo2::SetMarkerCount(marker_no);   }

// get set with specified ID
HaplotypeSet * HaplotypeSets::GetSet(const char* sid)
   {
   HaplotypeSet * set = FindSet(sid);
   set->setid = sid;
   return set;
   }

void * HaplotypeSets::create_set()
   {
   return new HaplotypeSet;
   }

HaplotypeSets::~HaplotypeSets()
   {
   for (int i = 0; i < Length(); i++)
      delete (HaplotypeSet *) objects[i];
   }

void HaplotypeSets::ShowCounts()
   {
   int totalHaplotypes = 0;
   int totalSets = 0;
   int knownHaplotypes = 0;
   int knownSets = 0;

   for (int i = Length() - 1; i >= 0; i--)
      {
      HaplotypeSet * set = (HaplotypeSet *) objects[i];

      if (set->Complexity() == 1.0)
         {
         knownHaplotypes += set->haploCount;
         knownSets++;
         }

      totalHaplotypes += set->haploCount;
      totalSets++;
      }

   //printf("Total: %d Haplotypes in %d Sets\n"
   //       "Known: %d Haplotypes in %d Sets [UNDERESTIMATE]\n\n",
   //       totalHaplotypes, totalSets,
   //       knownHaplotypes, knownSets);
   }

void HaplotypeSets::Filter(int maxBits)
   {
   int totalHaplotypes = 0;
   int totalSets = 0;
   int knownHaplotypes = 0;
   int knownSets = 0;

   //printf("Filtering data...\n");
   for (int i = Length() - 1; i >= 0; i--)
      {
      HaplotypeSet * set = (HaplotypeSet *) objects[i];

      double c = set->Complexity();
      double bits = log(c) / log(2.0);

      if (bits > maxBits)
         {
         //printf("DISCARDING family %s (%.1f bits)\n",
         //      (const char *) set->setid, bits);

         Delete(i);
         delete set;
         continue;
         }

      if (bits == 0.0)
         {
         knownHaplotypes += set->haploCount;
         knownSets++;
         }

      totalHaplotypes += set->haploCount;
      totalSets++;
      }

   //printf("\nTotal: %d Haplotypes in %d Sets\n"
   //       "Known: %d Haplotypes in %d Sets [UNDERESTIMATE]\n\n",
   //       totalHaplotypes, totalSets,
   //       knownHaplotypes, knownSets);
   }

void HaplotypeSets::UpdateAlleleCounts()
   {
   for (int i = 0; i < count; i++)
      ((HaplotypeSet *) objects[i])->UpdateAlleleCounts();
   }
 
