////////////////////////////////////////////////////////////////////// 
// fugue/Likelihood.cpp 
// (c) 2000-2003 Goncalo Abecasis
// 
// This file is distributed as part of the FUGUE source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile FUGUE.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Tuesday July 06, 2004
// 
 
#include "Likelihood.h"
#include "MathConstant.h"
#include "Random.h"
#include "Error.h"

#include <math.h>

void Likelihood::Initialize()
   {
   double number_of_haplotypes = 1;

   for (int i = 0; i < markers; i++)
      number_of_haplotypes *= alleleCounts[i];

   if (number_of_haplotypes > (1 << 26))
      error("The number of possible haplotypes (~2^%.1f) is too large",
             log(number_of_haplotypes) / log(2.0));

   haplos = (int) number_of_haplotypes;

   double memory = 3. * sizeof(double) * number_of_haplotypes * pow(2,-20);

   //printf("%d haplotype frequencies will be estimated\n"
   //       "  [~%.0f Mb of memory required]\n",
   //       haplos, memory);

   frequencies.Dimension(haplos);
   new_frequencies.Dimension(haplos);

   if (index != NULL) delete [] index;
   index = new IntArray[HaplotypeSet::maxHaploCount];

   init_llk = best_llk = 0;
   }

void Likelihood::SaveSummary()
   {
   if (best_frequencies.Length() == 0) return;

   FILE * f = fopen("first.freq", "at");
   fprintf(f, "%.5f\n", best_frequencies[0]);
   fclose(f);
   }

void Likelihood::Print(double minfreq)
   {
   String label, allele;

   //printf("Best log(lk) = %.2f\n\n", best_llk);

   //printf("Haplotypes with estimated frequency > %g\n", minfreq);

   //int count = 0;
   //double total = 0.0;

   for (int i = 0; i < best_frequencies.dim; i++)
      if (best_frequencies[i] > minfreq)
         {
         label = "";

         int haplo = i;
         for (int j = markers - 1; j >= 0; j--)
            {
            allele = haplo % alleleCounts[j] + 1;
            haplo /= alleleCounts[j];
            label  = allele + label;
            }

         printf("%6.2f%% %s\n", best_frequencies[i] * 100.0, (const char *) label);

         //total += best_frequencies[i];
         //count++;
         }

   //printf("\nThese %d haplotypes represent %.2f%% of total probability\n",
   //       count, total * 100);
   }

// get haplotype frequencies
void Likelihood::GetHapFreq(double hf[], double minfreq)
   {
   String label, allele, ref;
   bool initialize = false;

   for (int i = 0; i < best_frequencies.dim; i++)
      if (best_frequencies[i] > minfreq)
         {
         label = "";

         int haplo = i;
         for (int j = markers - 1; j >= 0; j--)
            {
            allele = haplo % alleleCounts[j] + 1;
            haplo /= alleleCounts[j];
            label  = allele + label;
            }

         if (!initialize) { ref = label; initialize = true; }
         int tmp1 = ref.SubStr(0,1).FastCompare(label.SubStr(0,1)) ? 0 : 2;
         int tmp2 = ref.SubStr(1,1).FastCompare(label.SubStr(1,1)) ? 0 : 1;
         hf[tmp1 + tmp2] = best_frequencies[i];
         }
   }

// perform LR test
double Likelihood::PerformTest()
   {
   return 2 * fabs(best_llk - init_llk);
   }

void Likelihood::RandomEM(HaplotypeSets * sets)
   {
   //printf("Starting with random allele frequencies...\n");
   for (int i = 0; i < frequencies.dim; i++)
      frequencies[i] = globalRandom.Next();

   frequencies /= frequencies.Sum();

   CoreEM(sets);
   //printf("\n\n");
  }

void Likelihood::EM(HaplotypeSets * sets)
   {
   //printf("Starting with equal allele frequencies...\n");
   frequencies.Set(1.0 / frequencies.dim);

   CoreEM(sets);
   //printf("\n\n");
   }

// estimate haplotype frequencies starting with
// founder allele frequencies (only for two markers)
void Likelihood::ModifiedEM(HaplotypeSets * sets, 
                       MarkerInfo * info1, MarkerInfo * info2)
   {
   frequencies.Zero();

   for (int i = 1; i < info1->freq.Length(); i++)
      {
      int a1 = (int) info1->GetAlleleLabel(i).AsInteger();
      double freq1 = info1->freq[i];
      for (int j = 1; j < info2->freq.Length(); j++)
         {
          int a2 = (int) info2->GetAlleleLabel(j).AsInteger();
          double freq2 = info2->freq[j];
          if (freq1 > 1e-7 && freq2 > 1e-7)
             frequencies[alleleCounts[1] * (a1 - 1) + a2 - 1] = freq1 * freq2;
         }
     }

   CoreEM(sets);
  }

void Likelihood::EditedEM(HaplotypeSets * sets)
   {
   int removed = 0;

   //printf("Removing very rare haplotypes...\n");
   for (int i = 0; i < frequencies.dim; i++)
      if (frequencies[i] < 1e-4)
         {
         frequencies[i] = 0;
         removed++;
         }
   //printf("Removed %d haplotypes\n\n", removed);

   frequencies /= frequencies.Sum();

   CoreEM(sets);
   //printf("\n\n");
   }

void Likelihood::CoreEM(HaplotypeSets * sets)
   {
   int count = 0;
   for (int i = 0; i < sets->Length(); i++)
      {
      HaplotypeSet * set = (HaplotypeSet *) sets->Object(i);
      count += set->haploCount;
      set->StandardizeWeights();
      }

   if (count == 0) return;

   double llk0 = 0.0, llk1;
   int    pass = 0;

   while (true)
      {
      llk1 = 0.0;

      new_frequencies.Zero();

      for (int i = 0; i < sets->Length(); i++)
         {
         HaplotypeSet * set = (HaplotypeSet *) sets->Object(i);

         secondPass = false;
         setLikelihood = SetLikelihood(set) * count;

         secondPass = true;
         SetLikelihood(set);

         llk1 += log(setLikelihood / count);
         }

      frequencies = new_frequencies;
      if (++pass == 1 && init_llk == 0) init_llk = llk1;

      double precision = max(fabs(llk1 + llk0) * tol, ztol);

      if (fabs(llk1 - llk0) < precision)
         break;

      llk0 = llk1;
      }

   if (best_llk == 0.0 || llk1 > best_llk)
      {
      best_llk = llk1;
      best_frequencies = frequencies;
      }
   }

double Likelihood::SetLikelihood(HaplotypeSet * set)
   {
   double lk = 0.0;

   for (int i = 0; i < set->graphCount; i++)
      lk += GraphLikelihood(set->graphs[i]);

   return lk;
   }

double Likelihood::GraphLikelihood(HaplotypeGraph * graph)
   {
   last.Dimension(graph->count);
   last.Set(0);

   marginals.Dimension(graph->count);

   for (int i = 0; i < graph->count; i++)
      {
      index[i].Dimension(1);
      index[i][0] = 0;

      UpdateIndex(graph, i);
      }

   return GraphLikelihood(graph, 0);
   }

double Likelihood::GraphLikelihood(HaplotypeGraph * graph, int i)
   {
   if (i == graph->unknowns.Length())
      return CurrentLikelihood(graph);

   Unknown * u = graph->unknowns.GetUnknown(i);

   if (u->type == U_MISSING)
      {
      MissingAllele * m = (MissingAllele *) u;

      if (graph->lastGenotype[m->haplotype] != -1)
         ExpandIndex(graph, m->haplotype);

      return GraphLikelihood(graph, i + 1);
      }

   /*** u->type == U_GRAPH ***/
   AlleleGraph * g = (AlleleGraph *) u;

   IntArray len(graph->count);
   IntArray pos = last;

   for (int j = 0; j < graph->count; j++)
      len[j] = index[j].Length();

   for (int j = 0; j < g->haplotypes.Length(); j++)
      {
      index[g->haplotypes[j]] *= alleleCounts[g->marker];
      index[g->haplotypes[j]] += g->alleles[j * 2] - 1;
      last[g->haplotypes[j]]++;
      UpdateIndex(graph, g->haplotypes[j]);
      }
   double lk = GraphLikelihood(graph, i + 1);

   for (int j = 0; j < graph->count; j++)
      {
      index[j].Dimension(len[j]);
      ReverseIndex(graph, j, pos[j]);
      }

   for (int j = 0; j < g->haplotypes.Length(); j++)
      {
      index[g->haplotypes[j]] *= alleleCounts[g->marker];
      index[g->haplotypes[j]] += g->alleles[j * 2 + 1] - 1;
      last[g->haplotypes[j]]++;
      UpdateIndex(graph, g->haplotypes[j]);
      }
   return lk + GraphLikelihood(graph, i + 1);
   }

double Likelihood::CurrentLikelihood(HaplotypeGraph * graph)
   {
   for (int i = 0; i < graph->count; i++)
      if (last[i] == 0)
         marginals[i] = 1.0;
      else if (last[i] == markers)
         {
         marginals[i] = frequencies[index[i][0]];

         for (int j = 1; j < index[i].Length(); j++)
            marginals[i] += frequencies[index[i][j]];
         }
      else
         error("Unexpected state while processing haplotype graph");

   double likelihood = graph->weight;

   for (int i = 0; i < graph->count; i++)
      likelihood *= marginals[i];

   if (secondPass && ((likelihood / setLikelihood) > 1e-7))
      {
      for (int i = 0; i < graph->count; i++)
         {
         double scale = likelihood / (setLikelihood * marginals[i]);

         if (last[i] == 0)
            for (int j = 0; j < new_frequencies.dim; j++)
               new_frequencies[j] += frequencies[j] * scale;
         else
            for (int j = 0; j < index[i].Length(); j++)
               new_frequencies[index[i][j]] += frequencies[index[i][j]] * scale;
         }

      return likelihood;
      }
   else if (secondPass)
      return 0.0;

   return likelihood;
   }

void Likelihood::UpdateIndex(HaplotypeGraph * graph, int i)
   {
   while (last[i] < markers && graph->haplotypes[i][last[i]] != 0)
      {
      index[i] *= alleleCounts[last[i]];
      index[i] += graph->haplotypes[i][last[i]] - 1;
      last[i]++;
      }
   }

void Likelihood::ReverseIndex(HaplotypeGraph * graph, int i, int pos)
   {
   if (pos == last[i]) return;

   int factor = alleleCounts[--last[i]];

   while (last[i] > pos)
      factor *= alleleCounts[--last[i]];

   index[i] /= factor;
   }

void Likelihood::ExpandIndex(HaplotypeGraph * graph, int i)
   {
   int factor = alleleCounts[last[i]];
   int original = index[i].Length();

   index[i] *= factor;
   index[i].Dimension(index[i].Length() * factor);

   for (int j = 1; j < factor; j++)
      for (int k = original * j; k < original * (j + 1); k++)
         index[i][k] = index[i][k - original] + 1;

   last[i]++;

   UpdateIndex(graph, i);
   }

void Likelihood::PrintExtras()
   {
   // Calculate information for this distribution
   double info = 0;

   for (int i = 0; i < best_frequencies.dim; i++)
      if (frequencies[i] > 1e-7)
         info += best_frequencies[i] * log(best_frequencies[i]);

   // Calculate D' between last marker and haplotype of other markers
   double D = 0.0;

   marginals.Dimension( alleleCounts[markers - 1] );
   marginals.Zero();

   int alleles = alleleCounts[markers - 1];

   for (int i = 0; i < best_frequencies.dim; i++)
      if (frequencies[i] > 1e-7)
         marginals[i % alleles] += frequencies[i];

   for (int i = 0; i < best_frequencies.dim; i += alleles)
      {
      double freq = 0.0;

      for (int j = 0; j < alleles; j++)
         freq += best_frequencies[i + j];

      if (freq < 1e-7) continue;

      for (int j = 0; j < alleles; j++)
         {
         double _D = best_frequencies[i + j] - freq * marginals[j];

         if (fabs(_D) < 1e-7) continue;

         double _Dmax = _D < 0 ?
                min(freq * marginals[j], (1. - freq) * (1. - marginals[j])) :
                min((1. - freq) * marginals[j], freq * (1. - marginals[j]));

         D += fabs(_D/_Dmax) * freq * marginals[j];
         }
      }

   // Calculate cumulative frequencies
   frequencies = best_frequencies;
   frequencies.Sort();

   double top1 = frequencies.dim > 0 ? frequencies[frequencies.dim - 1] : 1.0;
   double top2 = frequencies.dim > 1 ? frequencies[frequencies.dim - 2] + top1 : 1.0;
   double top3 = frequencies.dim > 2 ? frequencies[frequencies.dim - 3] + top2 : 1.0;
   double top4 = frequencies.dim > 3 ? frequencies[frequencies.dim - 4] + top3 : 1.0;
   double top5 = frequencies.dim > 4 ? frequencies[frequencies.dim - 5] + top4 : 1.0;

   printf("\n\nAdditional Information\n");
   printf("%10s %10s %10s %10s %10s %10s %10s\n",
          "Info", "D(last,*)", "Top1", "Top2", "Top3", "Top4", "Top5");
   printf("%10.3f %10.3f %10.3f %10.3f %10.3f %10.3f %10.3f\n",
          info, D, top1, top2, top3, top4, top5);
   }

void Likelihood::PrintTwoMarker()
   {
   printf("\n");

   if (markers != 2 || alleleCounts[0] != 2 || alleleCounts[1] != 2)
      {
      printf("Can only report two-marker ld for bi-allelic markers\n");
      return;
      }

  double p = best_frequencies[0] + best_frequencies[2];
  double r = best_frequencies[0] + best_frequencies[1];

  if (min(p, 1 - p) < 1e-8 || min(r, 1 - r) < 1e-8)
     {
     printf("One of the markers is monomorphic\n");
     return;
     }

  double D = best_frequencies[0] - p * r;
  double dprime = 0;

  if (D > 1e-8)
     dprime = D / min((1-p)*r, (1-r)*p);
  else if (D < -1e-8)
     dprime = D / min((1-p)*(1-r), p*r);

  dprime = fabs(dprime);

  double deltasq = (best_frequencies[0] * best_frequencies[3] -
                    best_frequencies[1] * best_frequencies[2]);
  deltasq *= deltasq;
  deltasq /= p * (1-p) * r * (1-r);

  printf("DeltaSq = %7.5f   Dprime = %7.5f\n", deltasq, dprime);
  }

 
